/**
* ProjectAES is a command-line implementation of the Advanced Encryption Standard(AES) algorithm.* 
*
* This project is based on the details & examples provided at:
*       http://csrc.nist.gov/groups/ST/toolkit/block_ciphers.html  and 
*       http://www.progressive-coding.com/tutorial.php?id=0
*
* This program is made freely available under the liberal MIT license. 
*  Please see the accompanying file, License.txt, for full details.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

// Program version number
#define VERSION "0.1"

// Values for VERBOSE level are 0 - 3 (3 being most detail)
#define VERBOSE 0

// Careful with this one. The key expander generates a lot of output when this is enabled
//#define DEBUG_KEY_EXPANDER

// Macros to make code a little cleaner
#define byte unsigned char
#define Nk Nkeys[currentMode]
#define Nb Nblocks[currentMode]
#define Nr Nrounds[currentMode]

// Constants
#define NUM_BYTES_IN_WORD 4
#define BLOCK_SIZE NUM_BYTES_IN_WORD * Nb

/**
* Global variables
**/
typedef enum {
    AES128 = 0,
    AES192,
    AES256
} MODE;
char NAMES[][8] = {
    "AES-128",
    "AES-192",
    "AES-256"
};
int currentMode = AES128; // Can be AES128, AES192, or AES256. Default is AES128.
int Nkeys[] = {4, 6, 8}; // Key length. Can be 4, 6 or 8 and reflects the number of columns in the Cipher Key
int Nblocks[] = {4, 4, 4}; // Block size, represents the number of WORDS in a block
int Nrounds[] = {10, 12, 14}; // Number of rounds. Can be 0x10, 0x12, or 14 depending on the keysize

byte MagicNumber[2] = {0x09, 0xf9}; // Numbers used in the header to identify the file type

// Precalculated non-linear transformation values
byte SBox[256] = {
0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16
};

byte InvSBox[256] = {
0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e, 0x81, 0xf3, 0xd7, 0xfb,
0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87, 0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb,
0x54, 0x7b, 0x94, 0x32, 0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e,
0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49, 0x6d, 0x8b, 0xd1, 0x25,
0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92,
0x6c, 0x70, 0x48, 0x50, 0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84,
0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05, 0xb8, 0xb3, 0x45, 0x06,
0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02, 0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b,
0x3a, 0x91, 0x11, 0x41, 0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73,
0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8, 0x1c, 0x75, 0xdf, 0x6e,
0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89, 0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b,
0xfc, 0x56, 0x3e, 0x4b, 0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4,
0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xec, 0x5f,
0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d, 0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef,
0xa0, 0xe0, 0x3b, 0x4d, 0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61,
0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d
};

byte Rcon[255] = {
0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a,
0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39,
0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a,
0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8,
0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef,
0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc,
0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b,
0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3,
0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94,
0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20,
0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35,
0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f,
0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04,
0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63,
0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd,
0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb
};

/**
*  Log table using 0xb7 as the generator 
*  Additional log tables may be found at:
*  http://www.samiam.org/galois.html
**/
byte LogTable[256] = {
0x00, 0xff, 0xc8, 0x3b, 0x91, 0x76, 0x04, 0xcf, 0x5a, 0x0b, 0x3f, 0x10, 0xcc, 0x11, 0x98, 0xb1,
0x23, 0xec, 0xd3, 0x3d, 0x08, 0x9f, 0xd8, 0x4c, 0x95, 0x25, 0xd9, 0x46, 0x61, 0x4b, 0x7a, 0xa7,
0xeb, 0xe2, 0xb5, 0xe0, 0x9c, 0xcd, 0x06, 0x86, 0xd0, 0xc0, 0x68, 0x60, 0xa1, 0x81, 0x15, 0xc3,
0x5e, 0xdf, 0xed, 0x28, 0xa2, 0x78, 0x0f, 0x54, 0x2a, 0x87, 0x14, 0xf6, 0x43, 0x03, 0x70, 0xda,
0xb4, 0x16, 0xab, 0xba, 0x7e, 0x20, 0xa9, 0x3e, 0x65, 0x49, 0x96, 0xc2, 0xce, 0x32, 0x4f, 0xf4,
0x99, 0x22, 0x89, 0x1b, 0x31, 0x63, 0x29, 0xac, 0x6a, 0x8f, 0x4a, 0x2b, 0xdd, 0x77, 0x8c, 0xb3,
0x27, 0x73, 0xa8, 0x1e, 0xb6, 0x1c, 0xf0, 0x45, 0x6b, 0xc1, 0x41, 0x6f, 0xd7, 0xc5, 0x1d, 0x09,
0xf2, 0xfe, 0x50, 0xdb, 0xdc, 0xc9, 0xbf, 0xbc, 0x0c, 0x0d, 0xcb, 0xfb, 0x39, 0x9b, 0xa3, 0x21,
0x7d, 0x5c, 0xde, 0x82, 0x74, 0x24, 0x83, 0xd6, 0x47, 0x42, 0xe8, 0x48, 0x72, 0x37, 0x07, 0xad,
0x2e, 0xbe, 0x12, 0x3a, 0x5f, 0x17, 0x8b, 0xaf, 0x97, 0xf7, 0xfa, 0xb8, 0x18, 0x1a, 0xbd, 0x05,
0x62, 0xc6, 0xea, 0xae, 0x52, 0x59, 0xe3, 0x9a, 0xf9, 0x80, 0x2c, 0x69, 0xf1, 0x4d, 0x75, 0x57,
0x33, 0x44, 0x58, 0xd2, 0x13, 0x36, 0xf3, 0x01, 0xa6, 0xf8, 0x40, 0xfc, 0x55, 0xaa, 0x7c, 0x53,
0xef, 0x2f, 0x3c, 0x51, 0x71, 0xf5, 0xe6, 0x4e, 0x7f, 0x79, 0xe4, 0x6e, 0xb9, 0x93, 0x0e, 0x5b,
0x34, 0x30, 0x8a, 0xe1, 0x0a, 0x7b, 0x38, 0x6d, 0xa0, 0xb7, 0x8e, 0x84, 0xe5, 0xfd, 0xd1, 0x90,
0xbb, 0xee, 0xc7, 0x35, 0x19, 0x67, 0xa4, 0xb2, 0xa5, 0xb0, 0x92, 0xca, 0x88, 0x66, 0x85, 0x2d,
0xd4, 0x1f, 0xd5, 0x5d, 0x94, 0x56, 0xc4, 0x8d, 0x02, 0xe7, 0x64, 0x9d, 0x6c, 0x26, 0xe9, 0x9e,
};

byte AntilogTable[256] = {
0x01, 0xb7, 0xf8, 0x3d, 0x06, 0x9f, 0x26, 0x8e, 0x14, 0x6f, 0xd4, 0x09, 0x78, 0x79, 0xce, 0x36,
0x0b, 0x0d, 0x92, 0xb4, 0x3a, 0x2e, 0x41, 0x95, 0x9c, 0xe4, 0x9d, 0x53, 0x65, 0x6e, 0x63, 0xf1,
0x45, 0x7f, 0x51, 0x10, 0x85, 0x19, 0xfd, 0x60, 0x33, 0x56, 0x38, 0x5b, 0xaa, 0xef, 0x90, 0xc1,
0xd1, 0x54, 0x4d, 0xb0, 0xd0, 0xe3, 0xb5, 0x8d, 0xd6, 0x7c, 0x93, 0x03, 0xc2, 0x13, 0x47, 0x0a,
0xba, 0x6a, 0x89, 0x3c, 0xb1, 0x67, 0x1b, 0x88, 0x8b, 0x49, 0x5a, 0x1d, 0x17, 0xad, 0xc7, 0x4e,
0x72, 0xc3, 0xa4, 0xbf, 0x37, 0xbc, 0xf5, 0xaf, 0xb2, 0xa5, 0x08, 0xcf, 0x81, 0xf3, 0x30, 0x94,
0x2b, 0x1c, 0xa0, 0x55, 0xfa, 0x48, 0xed, 0xe5, 0x2a, 0xab, 0x58, 0x68, 0xfc, 0xd7, 0xcb, 0x6b,
0x3e, 0xc4, 0x8c, 0x61, 0x84, 0xae, 0x05, 0x5d, 0x35, 0xc9, 0x1e, 0xd5, 0xbe, 0x80, 0x44, 0xc8,
0xa9, 0x2d, 0x83, 0x86, 0xdb, 0xee, 0x27, 0x39, 0xec, 0x52, 0xd2, 0x96, 0x5e, 0xf7, 0xda, 0x59,
0xdf, 0x04, 0xea, 0xcd, 0xf4, 0x18, 0x4a, 0x98, 0x0e, 0x50, 0xa7, 0x7d, 0x24, 0xfb, 0xff, 0x15,
0xd8, 0x2c, 0x34, 0x7e, 0xe6, 0xe8, 0xb8, 0x1f, 0x62, 0x46, 0xbd, 0x42, 0x57, 0x8f, 0xa3, 0x97,
0xe9, 0x0f, 0xe7, 0x5f, 0x40, 0x22, 0x64, 0xd9, 0x9b, 0xcc, 0x43, 0xe0, 0x77, 0x9e, 0x91, 0x76,
0x29, 0x69, 0x4b, 0x2f, 0xf6, 0x6d, 0xa1, 0xe2, 0x02, 0x75, 0xeb, 0x7a, 0x0c, 0x25, 0x4c, 0x07,
0x28, 0xde, 0xb3, 0x12, 0xf0, 0xf2, 0x87, 0x6c, 0x16, 0x1a, 0x3f, 0x73, 0x74, 0x5c, 0x82, 0x31,
0x23, 0xd3, 0x21, 0xa6, 0xca, 0xdc, 0xc6, 0xf9, 0x8a, 0xfe, 0xa2, 0x20, 0x11, 0x32, 0xe1, 0xc0,
0x66, 0xac, 0x70, 0xb6, 0x4f, 0xc5, 0x3b, 0x99, 0xb9, 0xa8, 0x9a, 0x7b, 0xbb, 0xdd, 0x71, 0x01,
};

// Function Prototypes
void AddRoundKey(byte state[], byte key[]);
void SubBytes(byte state[]);
void InvSubBytes(byte state[]);
void ShiftRows(byte state[]);
void InvShiftRows(byte state[]);
void MixColumns(byte state[]);
void InvMixColumns(byte state[]);
void MixColumn(byte column[]);
void InvMixColumn(byte column[]);
void RotWord(byte w[]);
void SubWord(byte w[]);
void KeyExpansion(byte key[], byte w[]);
long fileSize(FILE* ifp);
