/**
* ProjectAES is a command-line implementation of the Advanced Encryption Standard (AES) algorithm.
*
* This project is based on the details & examples provided at:
*       http://csrc.nist.gov/groups/ST/toolkit/block_ciphers.html
*       http://en.wikipedia.org/wiki/Advanced_Encryption_Standard
*       http://www.progressive-coding.com/tutorial.php?id=0
*
* This program is made freely available under the liberal MIT license. 
*  Please see the accompanying file, License.txt, for full details.
*/

#include "aes.h"
#include "random.c"
#include "aesTransformations.c"
#include "aesCore.c"
#include "aesErrors.c"
#include "aesUnitTests.c"

/**
* Returns the minimum of two numbers
**/
int min(int a, int b) {
    if(a < b) return a;
    return b;
}

/**
* Overwrite the contents of an array with zeros
**/
void ZeroArray(byte array[], int size){
    int i;
    for(i = 0; i < size; i++) {
        array[i] = 0;
    }
}

/**
* Returns the size of the file without destroying the current position in the file stream.
**/
long fileSize(FILE* ifp){
    long size;
    fpos_t pos; // current position in the stream
    
    // Save the current position in the stream so we can restore it later
    fgetpos(ifp, &pos);
    
    fseek(ifp, 0, SEEK_END);
	size = ftell(ifp);
	
    // Restore the position in the stream
    fsetpos(ifp, &pos);
    
    return size;
}

/**
* XOR two arrays and save the results in the first one. 
* Size is the length of the smaller array.
**/
void xor(byte a[], const byte b[], int size){
    int i;
    for(i = 0; i < size; i++) {
        a[i] ^= b[i];
    }
}

/**
* The function that performs the encryption process on a file.
* We will be using Cipher Block Chaining (CBC) with a unique Initialization Vector.
**/
void EncryptFile(FILE* ifp, FILE* ofp, byte key[], byte iv[]){
    long fileLength, i;
    byte w[BLOCK_SIZE*(Nr+1)];// Array holds the expanded key
    byte in[BLOCK_SIZE];      // Array holds one block of unencrypted data
    byte out[BLOCK_SIZE];     // Array holds one block of encrypted data
    
    #if VERBOSE >= 1
        printf("\nAES-");
        if(currentMode == AES128) printf("128");
        else if(currentMode == AES192) printf("192");
        else if(currentMode == AES256) printf("256");
        else printf("?");
        printf(" (Nk=%d, Nr=%d)\n", Nk, Nr);
        printf("CIPHER (ENCRYPT):\n");
    #endif
    
    // Find out how large the file is
	fileLength = fileSize(ifp);
    #if VERBOSE >= 1
        printf("The file is %ld bytes long\n", fileLength);
    #endif
    
    // Write the filesize to the header
    fwrite(&fileLength, 1, sizeof(fileLength), ofp);
    
    // Run the key expander
    KeyExpansion(key, w);
    
    // For each block
    for(i = 0; i < fileLength; i += BLOCK_SIZE){
        // read a single block
        fread(&in, 1, BLOCK_SIZE, ifp);
        
        #if VERBOSE >= 2
            printf("block[%2ld].input\t\t", i/(BLOCK_SIZE));
            PrintState(in);
        #endif
        
        // Perform Cipher-Block-Chaining on the cipher text
        if(i == 0) // First block
            xor(in, iv, BLOCK_SIZE); // XOR the IV with the plaintext
        else // Subsequent blocks
            xor(in, out, BLOCK_SIZE); // XOR the plaintext with the previous ciphertext
        
        // Run the cipher on the file block by block
        Cipher(in, out, w);
        
        #if VERBOSE >= 2
            printf("block[%2ld].output\t", i/(BLOCK_SIZE));
            PrintState(out);
        #endif
        
        // write a block
        fwrite(&out, 1, BLOCK_SIZE, ofp);
    }
    
    // Zero memory
    ZeroArray(w, BLOCK_SIZE*(Nr+1)); // Zero the expanded key
    ZeroArray(in, BLOCK_SIZE); // Zero the plaintext block
}

/**
* The function that performs the encryption process on a file.
* We will be using Cipher Block Chaining (CBC) with a unique Initialization Vector.
**/
void DecryptFile(FILE* ifp, FILE* ofp, byte key[], byte iv[]){
    long fileLength, i;
    byte w[BLOCK_SIZE*(Nr+1)];// Array holds the expanded key
    byte in[BLOCK_SIZE];      // Array holds one block of encrypted data
    byte in_prev[BLOCK_SIZE]; // Array holds previous block of encrypted data
    byte out[BLOCK_SIZE];     // Array holds one block of unencrypted data
    
    #if VERBOSE >= 1
        printf("\nAES-");
        if(currentMode == AES128) printf("128");
        else if(currentMode == AES192) printf("192");
        else if(currentMode == AES256) printf("256");
        else printf("?");
        printf(" (Nk=%d, Nr=%d)\n", Nk, Nr);
        printf("CIPHER (DECRYPT):\n");
    #endif
    
    // Find out how large the file is
	fread(&fileLength, 1, sizeof(fileLength), ifp);
    
    #if VERBOSE >= 1
        printf("The file is %ld bytes long\n", fileLength);
    #endif
    
    // Run the key expander
    KeyExpansion(key, w);
    
    // For each block
    for(i = 0; i < fileLength; i += BLOCK_SIZE){
        // read a single block
        fread(&in, 1, BLOCK_SIZE, ifp);
        
        #if VERBOSE >= 2
            printf("block[%2ld].input\t\t", i/(BLOCK_SIZE));
            PrintState(in);
        #endif
        
        // Run the cipher on the file block by block
        InvCipher(in, out, w);
        
        if(i == 0) // First block
            xor(out, iv, BLOCK_SIZE); // XOR the IV with the output
        else
            xor(out, in_prev, BLOCK_SIZE); // XOR the output with the previous ciphertext
        
        #if VERBOSE >= 2
            printf("block[%2ld].output\t", i/(BLOCK_SIZE));
            PrintState(out);
        #endif
        
        // write a block
        if(i > fileLength - BLOCK_SIZE) // The very last block
            fwrite(&out, 1, fileLength % (BLOCK_SIZE), ofp);  // Leave off the extra padding bytes
        else
            fwrite(&out, 1, BLOCK_SIZE, ofp); // Else write the full block
        
        // Save the ciphertext 
        memcpy(in_prev, in, BLOCK_SIZE);
    }
    
    // Zero memory
    ZeroArray(w, BLOCK_SIZE*(Nr+1));  // Zero the expanded key
    ZeroArray(out, BLOCK_SIZE);       // Zero the plaintext block
}

/**
* Initialize the Initialization Vector using a pseudo-random stream
* generated by RC4 using the current time as the key.
* Also writes the seed to the beginning of the file
**/
void initIV(byte iv[], FILE* ofp){
    int i;
    time_t rawtime;
    struct tm* timeinfo;

    time ( &rawtime );  // Get the current system time
    timeinfo = gmtime( &rawtime ); // Format it into GMT time
    
    // Write the seed to the output file
    fwrite(&rawtime, 1, 4, ofp);
    
    // Initialize the pseduo-random number generator using the ascii version of the date/time as the seed
    Random(asctime(timeinfo), strlen(asctime(timeinfo))); 
    
    // For each byte in the IV:
    for(i = 0; i < BLOCK_SIZE; i++) {
        iv[i] = RandomByte(); // Initialize to a random value
    }
    #if VERBOSE >= 1
        printf("IV: ");
        PrintState(iv);
    #endif
}
/**
* Get the seed for the IV from the input file
**/
void getIV(byte iv[], FILE* ifp){
    int i;
    time_t rawtime;
    struct tm * timeinfo;

    // Read seed from the input file
    fread(&rawtime, 1, 4, ifp);
    
    timeinfo = gmtime( &rawtime ); // Format time into GMT time
    
    // Initialize the pseduo-random number generator using the ascii version of the date/time as the seed
    Random(asctime(timeinfo), strlen(asctime(timeinfo)));
    
    // For each byte in the IV:
    for(i = 0; i < BLOCK_SIZE; i++) {
        iv[i] = RandomByte();
    }
    #if VERBOSE >= 1
        printf("IV: ");
        PrintState(iv);
    #endif
}

/**
* Show the user how to properly use the program
*/
void displayUsage(char* ProgName){
	printf("Usage: %s {MODE} -i{INPUT FILE} -o{OUTPUT FILE} {KEY} [{OPTIONS}]... \n\n", ProgName);
	printf("MODE (Required):\n");
    printf("  -e, --encrypt\t\ttoggle encryption mode\n");
	printf("  -d, --decrypt\t\ttoggle decryption mode\n");
    printf("\nKEY (Required):\n");
    printf("  -k \"{KEY}\"\t\tpassword for encryption\n");
    printf("  --keyfile {KEY FILE}\tuse a file as the key (Recomended)\n");
    printf("\n  NOTE: Both a key and keyfile may be used for extra security or to guard against keyloggers.\n");
    printf("\nOPTIONS:\n");
    printf("  --aes128 (default) \tuse AES-128. (Key must be 128 bits)\n");
    printf("  --aes192 \t\tuse AES-192. (Key must be 192 bits)\n");
    printf("  --aes256 \t\tuse AES-256. (Key must be 256 bits)\n");
    printf("  --test \t\trun diagnostic tests\n");
    printf("\nEXAMPLE:\n  Encrypt a file, in.txt, and save it as , out.txt:\n");
    printf("    >%s -e -i \"in.txt\" -o \"out.txt\" -k \"password123456789\"", ProgName);
    printf("\n");
}

/**
* MAIN - program entry point
*/
int main(int argc, char** argv){
    FILE* ifp = NULL;  // The input file pointer
    FILE* ofp = NULL;  // The output file pointer
    FILE* kfp = NULL;  // Optional key file pointer
    byte* key;  // The key array -- has to be large enough to hold the largest possible key (AES256)
    byte* iv;   // The IV is the same length as the key
    
    byte temp[2] = {0}; // A temporary byte array
    int i; // A counter
    int keyIndex = 0; // To avoid wasteful memory operations, we will simply keep track of the index in argv[] where the key appears
    
    // Flags
    char mode = 0;          // 1 = encryption, 2 = decryption
    char hasInputFile = 0;  // A flag to detect wether the user entred an input file
    char hasOutputFile = 0; // A flag to detect wether the user entred an output file
    char hasKey = 0;        // A flag to detect wether the user entred an encryption key
    char hasKeyFile = 0;    // A flag to detect wether the user entered a file to use as the ky
    
    // Print the title and description
    printf("ProjectAES %s, a command-line encryptor and decryptor.\n\n", VERSION);
    
    // Parse the parameters
	for(i = 1; i < argc; i++) {
        if(strcmp(argv[i], "-e") == 0 || strcmp(argv[i], "--encrypt") == 0) {
            mode = 1;
        }
        else if(strcmp(argv[i], "-d") == 0 || strcmp(argv[i], "--decrypt") == 0) {
            mode = 2;
        }
		else if(strcmp(argv[i], "-i") == 0) {
			ifp = fopen(argv[++i], "rb");
            
            if(ifp == NULL)
                fatalError(FILE_OPEN_ERROR, argv[i], NULL);
                
            hasInputFile = 1;
            #if VERBOSE >= 1
                printf("Input file: %s\n", argv[i]);
            #endif
		}
		else if(strcmp(argv[i], "-o") == 0) {
			ofp = fopen(argv[++i], "wb");
            
            if(ofp == NULL) 
                fatalError(FILE_OPEN_ERROR, argv[i], NULL);
                
            hasOutputFile = 1;
            #if VERBOSE >= 1
                printf("Output file: %s\n", argv[i]);
            #endif
		}
		else if(strcmp(argv[i], "-k") == 0) {
            hasKey = 1;
            keyIndex = ++i; // hasKey = the index where the key is found
		}
        else if(strcmp(argv[i], "--keyfile") == 0) {
            kfp = fopen(argv[++i], "rb");
            
            if(kfp == NULL) 
                fatalError(FILE_OPEN_ERROR, argv[i], NULL);
            
            hasKeyFile = 1;
            #if VERBOSE >= 1
                printf("Key file: %s\n", argv[i]);
            #endif
		}
		else if(strcmp(argv[i], "--aes128") == 0) {
			// Set the 'flavor' of encryption
            currentMode = AES128;
		}
        else if(strcmp(argv[i], "--aes192") == 0) {
			// Set the 'flavor' of encryption
            currentMode = AES192;
		}
        else if(strcmp(argv[i], "--aes256") == 0) {
			// Set the 'flavor' of encryption
            currentMode = AES256;
		}
        else if(argv[i][0] == '?' || strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "-?") == 0 || 
            strcmp(argv[i], "-help") == 0 || strcmp(argv[i], "--help") == 0){
            displayUsage(argv[0]);
            exit(0);
        }   
        else if(strcmp(argv[i], "--test") == 0){
            RunUnitTests();
            exit(0);
        }
        else {
            // Unknown flag
            printf("Error: unknown parameter \"%s\".", argv[i]);
            //displayUsage(argv[0]);
            exit(0);
        }
	}
    
    // Check that mandatory parameters have been filled in
	if(!mode || !hasInputFile || !hasOutputFile || !(hasKey || hasKeyFile) )
        fatalError(MISSING_PARAMETER, argv[0], NULL);
    
    // If we are decrypting, see what flavor of AES was used during the original encryption
    if(mode == 2){
        // Start by checking the magic number
        fread(&temp, 1, 2, ifp);
        if(memcmp(temp, MagicNumber, 2) != 0)
            fatalError(UNKOWN_FILE_TYPE, NULL, NULL);
    
        // Read the AES mode
        fread(&currentMode, 1, 1, ifp);
        #if VERBOSE >= 1
        printf("Encrypted file reports to be %s.\n", NAMES[currentMode]);
        #endif
    }
    
    // Allocate memory for the key now that we know which flavor of AES we are using
    key = (byte*) calloc(4*Nk, sizeof(byte));

    // Did the user want to use a key file?
    if(hasKeyFile){
        // Make sure the size matches the specs for requested 'flavor' of AES
        if( fileSize(kfp) < 4*Nk )
            fatalError(KEY_TOO_SHORT, NULL, NULL);
        else{
            // Copy the contents of the file to the key array
            fread(key, 1, 4*Nk, kfp);
            
            // User also wants to use a regular key too
            if(hasKey){
                // We XOR the keyfile and the password together to generate the encryption key
                xor(key, argv[keyIndex], min(4*Nk, strlen(argv[keyIndex])));
            }
            fclose(kfp);
        }
    }
    else{
        if( strlen(argv[keyIndex]) < 4*Nk )
            fatalError(KEY_TOO_SHORT, NULL, NULL);
            
        strncpy(key, argv[keyIndex], 4*Nk);
    }
    
    #if VERBOSE >= 2
        printf("Key: ");
        PrintKey(key);
    #endif
    
    // Allocate memory for the IV
    iv = (byte*) calloc(BLOCK_SIZE, sizeof(byte));
    
    // Peform encryption or decryption
    if(mode == 1){
        /** Build the header **/
        // Write the magic number
        fwrite(&MagicNumber, 1, 2, ofp);
        // Write the AES mode
        fwrite(&currentMode, 1, 1, ofp);
        
        // Initialize the IV and write the seed to the output file
        initIV(iv, ofp);
        
        // Now encrypt the file
        EncryptFile(ifp, ofp, key, iv);
        printf("Encryption successful.");
    }
    else {
        /** Retrieve the header **/
        // Mode was retrieved before we read in the key or keyfile
        
        // Initialize the IV using the seed sent with the file
        getIV(iv, ifp);
        
        DecryptFile(ifp, ofp, key, iv);
        printf("Decryption successful.");
    }
    
    /** Clean up **/
    // Close the open files
    fclose(ifp);
    fclose(ofp);
    
    // Zero the arrays before freeing them
    ZeroArray(key, 4*Nk);
    ZeroArray(iv, BLOCK_SIZE);
    
    // Free the memory
    free(key);
    free(iv);
	return 0;
}
