/**
* This file contains the four transformations for AES and their inverses:
*   AddRoundKey()
*   SubBytes()
*   ShiftRows()
*   MixColumns()
*
* This program is made freely available under the liberal MIT license. 
*  Please see the accompanying file, License.txt, for full details.
**/

/**
* Transformation in the Cipher and Inverse Cipher in which a Round
*  Key is added to the State using an XOR operation.  The length of a
*  Round Key equals the size of the State (i.e., for Nb = 4, the Round
*  Key length equals 128 bits/16 bytes).
**/
void AddRoundKey(byte state[], byte key[]) {
    int i;
    for(i = 0; i < 16; i++)
        state[i] ^= key[i];
}

/**
*  Transformation in the Cipher that processes the State using a non-
*  linear byte substitution table (S-box) that operates on each of the
*  State bytes independently.
**/
void SubBytes(byte state[]) {
    int i;
    
    for(i = 0; i < BLOCK_SIZE; i++) {
        // Replace each element in the array with a value from the Sbox
        state[i] = SBox[state[i]];
    }
}

/**
*  Transformation in the Inverse Cipher that is the inverse of
*  SubBytes().
**/
void InvSubBytes(byte state[]) {
    int i;
    
    for(i = 0; i < BLOCK_SIZE; i++) {
        // Replace each element in the array with a value from the Sbox
        state[i] = InvSBox[state[i]];
    }
}

/**
* Function used in the Key Expansion routine that takes a four-byte
* word [a0, a1, a2, a3], performs a cyclic permutation, and returns
* the word [a1, a2, a3, a0].
**/
void RotWord(byte w[]) {
    byte temp;
    int i;
    
    temp = w[0]; // temp = a0
    
    for(i=0; i < 3; i++)
        w[i] = w[i+1]; // a0 = a1, a1 = a2, a2 = a3
    
    w[3] = temp; // a3 = temp
}

/**
*  Transformation in the Cipher that processes the State by cyclically
*  shifting the last three rows of the State by different offsets.
**/
void ShiftRows(byte state[]) {
    int i, j, k;
    byte temp;
    
    // Iterate over the 4 rows
    for(i = 0; i < 4; i++)
    {
        for(j = 0; j < i; j++)
        {
            temp = state[i];
            
            for(k = 0; k < 3; k++){
                state[i + (k*4)] = state[i + (k+1)*4];
                //printf("state[%d] = state[%d]\n", i + (k*4), i + (k+1)*4);
            }
            state[12 + i] = temp;
            //printf("state[%d] = temp\n\n", 12 + i);
        }
    }
}

/**
*  Transformation in the Inverse Cipher that is the inverse of
*  ShiftRows().
**/
void InvShiftRows(byte state[]) {
    int i, j, k;
    byte temp;
    
    // Iterate over the 4 rows
    for(i = 0; i < 4; i++)
    {
        for(j = 0; j < i; j++)
        {
            temp = state[i + 12];
            //printf("\ntemp = state[%d]\n", i + 12);
            
            for(k = 2; k >= 0; k--){
                state[i + (k+1)*4] = state[i + k*4];
                //printf("state[%d] = state[%d]\n", i + (k+1)*4, i + k*4);
            }
            state[i] = temp;
            //printf("state[%d] = temp\n", i);
        }
    }
}
/**
byte GaloisMult(byte a, byte b){
    byte p = 0;  // product
    unsigned char counter;
    unsigned char hi_bit;
    
    for(counter = 0; counter < 8; counter++){
        if((b & 1) == 1) // if the low bit of 'b' is set
            p ^= a; // XOR the product by the value of 'a'
        hi_bit = a & 0x80;
        a <<= 1;  // Left shift 'a' by one
        if(hi_bit == 0x80) // if a's high bit was set prior to the rotation
            a ^= 0x1b;
        b >>= 1;  // Right shift 'b' by one
    }
    
    return p;
}
**/
// A faster version based on log tables
// Taken from http://www.samiam.org/galois.html
byte GaloisMult(byte a, byte b){
    int s;
    int q;
    int z = 0;
    
    s = LogTable[a] + LogTable[b];
    s %= 255;
    
    // Get the antilog value
    s = AntilogTable[s];
    
    /* Now, we have some fancy code that returns 0 if either
           a or b are zero; we write the code this way so that the
           code will (hopefully) run at a constant speed in order to
           minimize the risk of timing attacks */
	q = s;
	if(a == 0) {
		s = z;
	}
	else {
		s = q;
	}
	if(b == 0) {
		s = z;
	} 
	else {
		q = z;
	}
	return s;
}
/**
*  Transformation in the Cipher that takes all of the columns of the
*  State and mixes their data (independently of one another) to
*  produce new columns.
**/
void MixColumns(byte state[]) {
    int i;
    
    for(i = 0; i < Nb; i++) {
        MixColumn(state+(i*4));
    }
}

/**
* Mix a single column
**/
void MixColumn(byte column[]) {
    byte a[4];  // simply a copy of the input array 
    memcpy(a, column, 4);
    
    column[0] = GaloisMult(a[0], 2) ^ a[3] ^ a[2] ^ GaloisMult(a[1], 3);  //    r0 = 2a0 + a3 + a2 + 3a1
    column[1] = GaloisMult(a[1], 2) ^ a[0] ^ a[3] ^ GaloisMult(a[2], 3);  //    r1 = 2a1 + a0 + a3 + 3a2
    column[2] = GaloisMult(a[2], 2) ^ a[1] ^ a[0] ^ GaloisMult(a[3], 3);  //    r2 = 2a2 + a1 + a0 + 3a3
    column[3] = GaloisMult(a[3], 2) ^ a[2] ^ a[1] ^ GaloisMult(a[0], 3);  //    r3 = 2a3 + a2 + a1 + 3a0
}

/**
* Transformation in the Inverse Cipher that is the inverse of
*  MixColumns().
**/
void InvMixColumns(byte state[]) {
    int i;
    
    for(i = 0; i < Nb; i++) {
        InvMixColumn(state+(i*4));
    }
}

/**
* Inverse mix a single column
**/
void InvMixColumn(byte column[]) {
    byte a[4];  // simply a copy of the input array 
    memcpy(a, column, 4);
    
    column[0] = GaloisMult(a[0], 14) ^ GaloisMult(a[3], 9) ^ GaloisMult(a[2], 13) ^ GaloisMult(a[1], 11);  //    r0 = 2a0 + a3 + a2 + 3a1
    column[1] = GaloisMult(a[1], 14) ^ GaloisMult(a[0], 9) ^ GaloisMult(a[3], 13) ^ GaloisMult(a[2], 11);  //    r1 = 2a1 + a0 + a3 + 3a2
    column[2] = GaloisMult(a[2], 14) ^ GaloisMult(a[1], 9) ^ GaloisMult(a[0], 13) ^ GaloisMult(a[3], 11);  //    r2 = 2a2 + a1 + a0 + 3a3
    column[3] = GaloisMult(a[3], 14) ^ GaloisMult(a[2], 9) ^ GaloisMult(a[1], 13) ^ GaloisMult(a[0], 11);  //    r3 = 2a3 + a2 + a1 + 3a0
}

/** 
*  Function used in the Key Expansion routine that takes a four-byte
*  input word and applies an S-box to each of the four bytes to
*  produce an output word.
**/
void SubWord(byte w[]) {
    int i;
    
    for(i = 0; i < 4; i++) {
        // Replace each element in the array with a value from the Sbox
        w[i] = SBox[w[i]];
    }

}
