/**
* This is a cryptographic psuedo-random number generator based on RC4.
*
* Before RandomByte() will start returning random numbers, the random
* number generator must be initialized by calling Random() with a key.
*
* This program is made freely available under the liberal MIT license. 
*  Please see the accompanying file, License.txt, for full details.
**/

#include <stdio.h>

#define THROW_AWAY 4096
#define byte unsigned char

// Globals
byte PRNG_S[256];
byte PRNG_S2[256];
byte PRNG_init = 0; // A flag that gets set when the PRNG has been initialized
int PRNG_i, PRNG_j; // Two counters

// A default key -- to be used if the PRNG was not initalized
byte PRNG_default_key[16] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 
0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff};


/**
* Initialize the random number generator by passing it a key as the seed
**/
void Random(const byte* key, const int keylen) {    
    byte temp;
    int index;

    // Set the initial values for both S-boxes
	for(PRNG_i = 0; PRNG_i < 256; PRNG_i++) {
		PRNG_S[PRNG_i] = PRNG_i; // fill the first S-box
		PRNG_S2[PRNG_i] = key[PRNG_i % keylen]; // Initialze the second S-box 
	}
	
	// Initialize the S-box
	for(PRNG_i = 0; PRNG_i < 256; PRNG_i++) {
		PRNG_j = (PRNG_j + PRNG_S[PRNG_i] + PRNG_S2[PRNG_i]) % 256;
		temp = PRNG_S[PRNG_i];
		PRNG_S[PRNG_i] = PRNG_S[PRNG_j];
		PRNG_S[PRNG_j] = temp;
	}
	
	// Initialize S2 to zero to prevent snooping
	for(PRNG_i = 0; PRNG_i < 256; PRNG_i++)
		PRNG_S2[PRNG_i] = 0;
	
	// Begin the actual encryption 
	PRNG_i = PRNG_j = 0;
	
	// Discard first 4096 bytes
	for(index = 0; index < 4096; index++) {
		// Generate a pseduorandom byte K
		PRNG_i = (PRNG_i + 1) % 256;
		PRNG_j = (PRNG_j + PRNG_S[PRNG_i]) % 256;
		temp = PRNG_S[PRNG_i]; // swap S[i] and S[j]
		PRNG_S[PRNG_i] = PRNG_S[PRNG_j];
		PRNG_S[PRNG_j] = temp;
	}
    
    // Set the initalized byte
    PRNG_init = 1;
}

/**
* Returns a psuedo random stream of bytes one at a time.
* Do NOT use this until you have initialized the generator.
**/
byte RandomByte() {
    byte temp;
    
    // if generator has not been initialized
    if(PRNG_init != 1)
        Random(PRNG_default_key, 16);
    
    // Generate a pseduorandom byte 
    PRNG_i = (PRNG_i + 1) % 256;
    PRNG_j = (PRNG_j + PRNG_S[PRNG_i]) % 256;
    temp = PRNG_S[PRNG_i]; // swap S[i] and S[j]
    PRNG_S[PRNG_i] = PRNG_S[PRNG_j];
    PRNG_S[PRNG_j] = temp;
    
    return PRNG_S[(PRNG_S[PRNG_i] + PRNG_S[PRNG_j]) % 256];
}
